@@ -12,6 +12,7 @@ use crate::{
 use std::{
     cmp::Ordering::*,
     collections::{HashMap, HashSet},
+    default::Default,
     hash::Hash,
 };
 
@@ -48,7 +49,7 @@ use std::{
 /// sets.make_set(4).unwrap();
 /// sets.make_set(9).unwrap();
 ///
-/// sets.union(4, 9).unwrap();
+/// sets.union(&4, &9).unwrap();
 ///
 /// // `into_iter` returns an iterator whose item is of type `HashSet<T>`
 /// let as_vec: Vec<HashSet<usize>> = sets.into_iter().collect();
@@ -79,6 +80,31 @@ where
     T: Copy + Eq + Hash,
 {
     nodes: HashMap<T, Node<T>>,
+    strategy: FindStrategy,
+}
+
+/// Heuristics to use in order to improve the algorithmic complexity of the algorithm.
+///
+/// These heuristics will affect how `find_set` modifies the forest.
+///
+/// See the [Wikipedia article](https://en.wikipedia.org/wiki/Disjoint-set_data_structure) for more
+/// information about these heuristics.
+#[derive(Clone, Copy, Debug, Eq, PartialEq)]
+pub enum FindStrategy {
+    /// during `find_set(x)`, for all nodes in the path from `x` to the root node, including `x`
+    /// but excluding the root node, set their parent to the root node
+    PathCompression,
+
+    /// during `find_set(x)`, for all nodes in the path from `x` to the root node, including `x`
+    /// but excluding the root node, set their parent to their grand parent
+    PathSplitting,
+}
+
+//
+impl Default for FindStrategy {
+    fn default() -> Self {
+        FindStrategy::PathCompression
+    }
 }
 
 //
@@ -89,7 +115,7 @@ where
     /// See [`UnionFind`](crate::traits::UnionFind) for an example.
     fn make_set(&mut self, item: T) -> Result<()> {
         //
-        match self.contains(item) {
+        match self.contains(&item) {
             // if `item` alreday exists, error with the approriate variant
             true => Err(Error::ItemAlreadyExists),
             // only make the set if it doesn't already exist
@@ -104,47 +130,27 @@ where
     }
 
     /// See [`UnionFind`](crate::traits::UnionFind) for an example.
-    fn union(&self, x: T, y: T) -> Result<()> {
+    fn union(&mut self, x: &T, y: &T) -> Result<()> {
         //
         match x == y {
             // if they are equivalent, no need to merge their sets
             true => Ok(()),
             // merge the two sets by first finding their representatives
             false => self.link(
-                self.find_set(x)?, //
-                self.find_set(y)?, //
+                &self.find_set(x)?, //
+                &self.find_set(y)?, //
             ),
         }
     }
 
     /// See [`UnionFind`](crate::traits::UnionFind) for an example.
-    fn find_set(&self, item: T) -> Result<T> {
+    fn find_set(&self, item: &T) -> Result<T> {
         //
-        match self.get_node(&item) {
-            // `item` exists in this `DisjointSets` and is stored in `node`
-            Some(node) => {
-                //
-                debug_assert!(item == node.get_item());
-
-                //
-                match node.get_parent() {
-                    // `node` is not a root node, so find the representative nad do path
-                    // compression
-                    Some(parent) => {
-                        // 1. find the representative by recursing
-                        let representative = self.find_set(parent)?;
-                        // 2. set the parent of `node` to the representative
-                        node.set_parent(representative);
-                        //
-                        Ok(representative)
-                    }
-                    // `node` is a root node, so return `item` since it's the representative of its
-                    // set
-                    None => Ok(item),
-                }
-            }
-            // `item` does not exist in this set
-            None => Err(Error::ItemDoesNotExist),
+        use crate::disjoint_sets::FindStrategy::*;
+        //
+        match self.strategy {
+            PathCompression => self.find_set_path_compression(item),
+            PathSplitting => self.find_set_path_splitting(item),
         }
     }
 }
@@ -166,6 +172,49 @@ where
     pub fn new() -> Self {
         Self {
             nodes: HashMap::new(),
+            strategy: Default::default(),
+        }
+    }
+
+    /// Creates an empty instance of `DisjointSets` with a specific strategy to be used by the
+    /// `find_set` operation.
+    ///
+    /// See [FindStrategy](crate::disjoint_sets::FindStrategy).
+    ///
+    /// # Example
+    ///
+    /// ```
+    /// use union_find_rs::prelude::*;
+    ///
+    /// let mut sets_1: DisjointSets<usize> = DisjointSets::with_strategy(FindStrategy::PathCompression);
+    ///
+    /// let mut sets_2: DisjointSets<usize> = DisjointSets::with_strategy(FindStrategy::PathSplitting);
+    /// ```
+    pub fn with_strategy(strategy: FindStrategy) -> Self {
+        Self {
+            nodes: HashMap::new(),
+            strategy,
+        }
+    }
+
+    /// Creates an empty instance of `DisjointSets` with a specific strategy to be used by the
+    /// `find_set` operation.
+    ///
+    /// See [FindStrategy](crate::disjoint_sets::FindStrategy).
+    ///
+    /// # Example
+    ///
+    /// ```
+    /// use union_find_rs::prelude::*;
+    ///
+    /// let mut sets_1: DisjointSets<usize> = DisjointSets::with_capacity_strategy(512, FindStrategy::PathCompression);
+    ///
+    /// let mut sets_2: DisjointSets<usize> = DisjointSets::with_capacity_strategy(512, FindStrategy::PathSplitting);
+    /// ```
+    pub fn with_capacity_strategy(capacity: usize, strategy: FindStrategy) -> Self {
+        Self {
+            nodes: HashMap::with_capacity(capacity),
+            strategy,
         }
     }
 
@@ -184,6 +233,7 @@ where
     pub fn with_capacity(capacity: usize) -> Self {
         Self {
             nodes: HashMap::with_capacity(capacity),
+            strategy: Default::default(),
         }
     }
 
@@ -215,11 +265,11 @@ where
     ///
     /// sets.make_set(9).unwrap();
     ///
-    /// assert_eq!(sets.contains(9), true);
-    /// assert_eq!(sets.contains(7), false);
+    /// assert_eq!(sets.contains(&9), true);
+    /// assert_eq!(sets.contains(&7), false);
     /// ```
-    pub fn contains(&self, item: T) -> bool {
-        self.nodes.contains_key(&item)
+    pub fn contains(&self, item: &T) -> bool {
+        self.nodes.contains_key(item)
     }
 
     /// Clears this `DisjointSets`, removing all data. Keeps the allocated memory for reuse.
@@ -301,12 +351,12 @@ where
     }
 
     // return the `Node` representation of `item`, if it exists
-    fn get_node(&self, item: &T) -> Option<&Node<T>> {
-        self.nodes.get(item)
+    fn get_node(&self, item: &T) -> Result<&Node<T>> {
+        self.nodes.get(item).ok_or(Error::ItemDoesNotExist)
     }
 
     // merge the two sets for which `x` and `y` are representatives
-    fn link(&self, x: T, y: T) -> Result<()> {
+    fn link(&self, x: &T, y: &T) -> Result<()> {
         // assert that `x` and `y` are indeed root nodes
         debug_assert!(self.get_node(&x).unwrap().get_parent().is_none());
         debug_assert!(self.get_node(&y).unwrap().get_parent().is_none());
@@ -326,12 +376,12 @@ where
                 // union-by-rank
                 match x_rank.cmp(&y_rank) {
                     // node with the larger rank becomes the parent
-                    Greater => y_node.set_parent(x),
-                    Less => x_node.set_parent(y),
+                    Greater => y_node.set_parent(&x_node),
+                    Less => x_node.set_parent(&y_node),
                     // if both ranks are equal, then choose the parent arbitrarily
                     // and increment the parent's rank by 1
                     Equal => {
-                        x_node.set_parent(y);
+                        x_node.set_parent(&y_node);
                         y_node.set_rank(y_rank + 1);
                     }
                 }
@@ -342,6 +392,71 @@ where
             false => Err(Error::ItemDoesNotExist),
         }
     }
+
+    // this method follows the semantics of the pseudo code more closely
+    //
+    // `Node::<T>::get_parent` returns `None` if the node is a root node, whereas in the pseudo
+    // code root nodes return themselves when their parents are queried
+    //
+    // this method thus returns `Ok(t)` when called on `t` and `t` is a root node, and returns an
+    // `Err` only if `t` does not exist in this forest
+    fn get_parent<'a>(&'a self, item: &'a Node<T>) -> Result<&'a Node<T>> {
+        match item.get_parent() {
+            //
+            Some(parent) => self.get_node(&parent),
+            // if `item` doesn't have a parent, return `item`
+            None => Ok(item),
+        }
+    }
+
+    // do `find_set` by doing full path compression
+    fn find_set_path_compression(&self, item: &T) -> Result<T> {
+        // `item` exists in this `DisjointSets` and is stored in `node`
+        let node = self.get_node(&item)?;
+        let parent = self.get_parent(node)?;
+
+        // check if `node` is a root node
+        match node == parent {
+            // `node` is a root node, so just return itself
+            true => Ok(node.get_item()),
+            // `node` is not a root node, so find the representative and do path
+            // compression
+            false => {
+                // 1. find the representative by recursing
+                let representative_item = self.find_set(&parent.get_item())?;
+                let representative = self.get_node(&representative_item)?;
+                // 2. set the parent of `node` to the representative
+                node.set_parent(&representative);
+                //
+                Ok(representative.get_item())
+            }
+        }
+    }
+
+    // do `find_set` by doing path splitting
+    fn find_set_path_splitting(&self, item: &T) -> Result<T> {
+        //
+        let node = self.get_node(&item)?;
+
+        // `item` exists in this `DisjointSets` and is stored in `node`
+        debug_assert!(item == &node.get_item());
+
+        //
+        let mut x = node;
+        let mut x_parent = self.get_parent(node)?;
+
+        // while `x` is not root
+        while x != x_parent {
+            //
+            let x_parent_parent = self.get_parent(x_parent)?;
+            if x_parent != x_parent_parent {
+                x_parent.set_parent(x_parent_parent);
+            }
+            x = x_parent;
+            x_parent = x_parent_parent;
+        }
+        Ok(x.get_item())
+    }
 }
 
 /// In order to allow easy access to the disjoint sets stored in the forest, this struct implements
@@ -356,7 +471,7 @@ where
 /// sets.make_set(4).unwrap();
 /// sets.make_set(9).unwrap();
 ///
-/// sets.union(4, 9).unwrap();
+/// sets.union(&4, &9).unwrap();
 ///
 /// // `into_iter` returns an iterator whose item is of type `HashSet<T>`
 /// let as_vec: Vec<HashSet<usize>> = sets.into_iter().collect();
@@ -384,7 +499,7 @@ where
             self.nodes
                 .iter()
                 .fold(HashMap::with_capacity(self.len()), |mut acc, (k, _)| {
-                    let parent = self.find_set(*k).unwrap();
+                    let parent = self.find_set(k).unwrap();
                     if !acc.contains_key(&parent) {
                         acc.insert(parent, HashSet::new());
                     }
@@ -411,25 +526,26 @@ mod tests {
         hash::{Hash, Hasher},
     };
 
+    // 2^16 = 65536
+    const ITEM_COUNT: usize = 1 << 16;
+
     // create a
     macro_rules! test_with_different_types {
         //
-        ( $mod_name:ident, $test_fn_name:ident ) => {
+        ( $mod_name:ident, $test_fn_name:ident, $disjoint_sets:expr ) => {
             //
             #[cfg(test)]
             mod $mod_name {
                 use super::*;
 
-                //
-                const ITEM_COUNT: usize = 4096;
-
                 //
                 #[test]
                 fn test_with_usize() {
                     //
                     let items: Vec<usize> = (0..ITEM_COUNT).collect();
+                    let dsets = $disjoint_sets;
                     //
-                    $test_fn_name(items);
+                    $test_fn_name(items, dsets);
                 }
 
                 //
@@ -438,8 +554,9 @@ mod tests {
                     //
                     const ITEM_COUNT_I64: i64 = ITEM_COUNT as i64;
                     let items: Vec<i64> = (-ITEM_COUNT_I64..ITEM_COUNT_I64).collect();
+                    let dsets = $disjoint_sets;
                     //
-                    $test_fn_name(items);
+                    $test_fn_name(items, dsets);
                 }
 
                 //
@@ -453,8 +570,9 @@ mod tests {
                     //
                     let items: Vec<CustomStruct> =
                         (0..ITEM_COUNT).map(|i| CustomStruct { id: i }).collect();
+                    let dsets = $disjoint_sets;
                     //
-                    $test_fn_name(items);
+                    $test_fn_name(items, dsets);
                 }
 
                 //
@@ -470,90 +588,69 @@ mod tests {
                         (0..ITEM_COUNT).map(|i| CustomStruct { id: i }).collect();
                     //
                     let refs: Vec<&CustomStruct> = items.iter().collect();
+                    let dsets = $disjoint_sets;
                     //
-                    $test_fn_name(refs);
+                    $test_fn_name(refs, dsets);
                 }
             }
         };
     }
 
     //
-    fn path_compression_tester<T>(items: Vec<T>)
+    fn path_compression_tester<T>(items: Vec<T>, mut dsets: DisjointSets<T>)
     where
         T: Copy + Eq + Hash,
     {
         //
-        let mut union_find = DisjointSets::new();
-        //
-        items
-            .iter()
-            .copied()
-            .for_each(|item| union_find.make_set(item).unwrap());
+        items.iter().for_each(|item| dsets.make_set(*item).unwrap());
 
         //
         items.iter().enumerate().skip(1).for_each(|(i, item)| {
-            let prev_item: T = *items.get(i - 1).unwrap();
-            union_find.union(*item, prev_item).unwrap();
+            let prev_item: &T = items.get(i - 1).unwrap();
+            dsets.union(item, prev_item).unwrap();
         });
 
         // check that all items have the same representative
         let reps: HashSet<_> = items
             .iter()
-            .copied()
-            .map(|item| union_find.find_set(item).unwrap())
+            .map(|item| dsets.find_set(item).unwrap())
             .collect();
         //
         assert_eq!(reps.len(), 1);
     }
 
     //
-    fn make_set_only_tester<T>(items: Vec<T>)
+    fn make_set_only_tester<T>(items: Vec<T>, mut dsets: DisjointSets<T>)
     where
         T: Copy + Eq + Hash,
     {
         //
-        let mut union_find = DisjointSets::new();
-        //
-        items
-            .iter()
-            .copied()
-            .for_each(|item| union_find.make_set(item).unwrap());
+        items.iter().for_each(|item| dsets.make_set(*item).unwrap());
 
         // check that all items have the same representative
         let reps: HashSet<_> = items
             .iter()
-            .copied()
-            .map(|item| union_find.find_set(item).unwrap())
+            .map(|item| dsets.find_set(item).unwrap())
             .collect();
         //
         assert_eq!(reps.len(), items.len());
     }
 
     //
-    fn union_all_tester<T>(items: Vec<T>)
+    fn union_all_tester<T>(items: Vec<T>, mut dsets: DisjointSets<T>)
     where
         T: Copy + Eq + Hash,
     {
         //
-        let mut union_find = DisjointSets::new();
-        //
-        items
-            .iter()
-            .copied()
-            .for_each(|item| union_find.make_set(item).unwrap());
+        items.iter().for_each(|item| dsets.make_set(*item).unwrap());
 
         //
-        items
-            .iter()
-            .copied()
-            .enumerate()
-            .skip(1)
-            .for_each(|(i, item)| {
-                union_find.union(item, *items.get(i - 1).unwrap()).unwrap();
-            });
+        items.iter().enumerate().skip(1).for_each(|(i, item)| {
+            dsets.union(item, items.get(i - 1).unwrap()).unwrap();
+        });
 
         // check that there is only 1 set
-        let sets: Vec<HashSet<T>> = union_find.into_iter().collect();
+        let sets: Vec<HashSet<T>> = dsets.into_iter().collect();
         assert_eq!(sets.len(), 1);
     }
 
@@ -567,33 +664,28 @@ mod tests {
         s.finish()
     }
     //
-    fn union_by_hash_parity_tester<T>(items: Vec<T>)
+    fn union_by_hash_parity_tester<T>(items: Vec<T>, mut dsets: DisjointSets<T>)
     where
         T: Copy + Eq + Hash,
     {
         //
-        let mut union_find = DisjointSets::new();
-        //
-        items
-            .iter()
-            .copied()
-            .for_each(|item| union_find.make_set(item).unwrap());
+        items.iter().for_each(|item| dsets.make_set(*item).unwrap());
 
         let mut hash_mod_even = None;
         let mut hash_mod_odd = None;
 
         //
-        items.iter().copied().for_each(|item| {
+        items.iter().for_each(|item| {
             if hash(&item) % 2 == 0 {
                 match hash_mod_even {
-                    Some(x) => union_find.union(item, x).unwrap(),
+                    Some(x) => dsets.union(item, x).unwrap(),
                     None => {
                         hash_mod_even = Some(item);
                     }
                 }
             } else {
                 match hash_mod_odd {
-                    Some(x) => union_find.union(item, x).unwrap(),
+                    Some(x) => dsets.union(item, x).unwrap(),
                     None => {
                         hash_mod_odd = Some(item);
                     }
@@ -602,15 +694,163 @@ mod tests {
         });
 
         // check that all items have the same representative
-        let sets: Vec<HashSet<T>> = union_find.into_iter().collect();
+        let sets: Vec<HashSet<T>> = dsets.into_iter().collect();
         //
         assert_eq!(sets.len(), 2);
         assert!(sets.iter().any(|set| set.iter().all(|x| hash(x) % 2 == 0)));
         assert!(sets.iter().any(|set| set.iter().all(|x| hash(x) % 2 == 1)));
     }
 
-    test_with_different_types!(make_set_only, make_set_only_tester);
-    test_with_different_types!(union_by_hash_parity, union_by_hash_parity_tester);
-    test_with_different_types!(path_compression, path_compression_tester);
-    test_with_different_types!(union_all, union_all_tester);
+    //
+    #[cfg(test)]
+    mod construct_with_new {
+        use super::*;
+        //
+        test_with_different_types!(make_set_only, make_set_only_tester, DisjointSets::new());
+        test_with_different_types!(
+            union_by_hash_parity,
+            union_by_hash_parity_tester,
+            DisjointSets::new()
+        );
+        test_with_different_types!(
+            path_compression,
+            path_compression_tester,
+            DisjointSets::new()
+        );
+        test_with_different_types!(union_all, union_all_tester, DisjointSets::new());
+    }
+
+    //
+    #[cfg(test)]
+    mod construct_with_capacity {
+        use super::*;
+        //
+        test_with_different_types!(
+            make_set_only,
+            make_set_only_tester,
+            DisjointSets::with_capacity(ITEM_COUNT)
+        );
+        test_with_different_types!(
+            union_by_hash_parity,
+            union_by_hash_parity_tester,
+            DisjointSets::with_capacity(ITEM_COUNT)
+        );
+        test_with_different_types!(
+            path_compression,
+            path_compression_tester,
+            DisjointSets::with_capacity(ITEM_COUNT)
+        );
+        test_with_different_types!(
+            union_all,
+            union_all_tester,
+            DisjointSets::with_capacity(ITEM_COUNT)
+        );
+    }
+
+    //
+    #[cfg(test)]
+    mod construct_with_path_compression {
+        use super::*;
+        //
+        //
+        test_with_different_types!(
+            make_set_only,
+            make_set_only_tester,
+            DisjointSets::with_strategy(FindStrategy::PathCompression)
+        );
+        test_with_different_types!(
+            union_by_hash_parity,
+            union_by_hash_parity_tester,
+            DisjointSets::with_strategy(FindStrategy::PathCompression)
+        );
+        test_with_different_types!(
+            path_compression,
+            path_compression_tester,
+            DisjointSets::with_strategy(FindStrategy::PathCompression)
+        );
+        test_with_different_types!(
+            union_all,
+            union_all_tester,
+            DisjointSets::with_strategy(FindStrategy::PathCompression)
+        );
+    }
+
+    //
+    #[cfg(test)]
+    mod construct_with_path_splitting {
+        use super::*;
+        //
+        //
+        test_with_different_types!(
+            make_set_only,
+            make_set_only_tester,
+            DisjointSets::with_strategy(FindStrategy::PathSplitting)
+        );
+        test_with_different_types!(
+            union_by_hash_parity,
+            union_by_hash_parity_tester,
+            DisjointSets::with_strategy(FindStrategy::PathSplitting)
+        );
+        test_with_different_types!(
+            union_all,
+            union_all_tester,
+            DisjointSets::with_strategy(FindStrategy::PathSplitting)
+        );
+    }
+
+    //
+    #[cfg(test)]
+    mod construct_with_capacity_path_compression {
+        use super::*;
+        //
+        //
+        test_with_different_types!(
+            make_set_only,
+            make_set_only_tester,
+            DisjointSets::with_capacity_strategy(ITEM_COUNT, FindStrategy::PathCompression)
+        );
+        test_with_different_types!(
+            union_by_hash_parity,
+            union_by_hash_parity_tester,
+            DisjointSets::with_capacity_strategy(ITEM_COUNT, FindStrategy::PathCompression)
+        );
+        test_with_different_types!(
+            path_compression,
+            path_compression_tester,
+            DisjointSets::with_capacity_strategy(ITEM_COUNT, FindStrategy::PathCompression)
+        );
+        test_with_different_types!(
+            union_all,
+            union_all_tester,
+            DisjointSets::with_capacity_strategy(ITEM_COUNT, FindStrategy::PathCompression)
+        );
+    }
+
+    //
+    #[cfg(test)]
+    mod construct_with_capacity_path_splitting {
+        use super::*;
+        //
+        //
+        test_with_different_types!(
+            make_set_only,
+            make_set_only_tester,
+            DisjointSets::with_capacity_strategy(ITEM_COUNT, FindStrategy::PathSplitting)
+        );
+        test_with_different_types!(
+            union_by_hash_parity,
+            union_by_hash_parity_tester,
+            DisjointSets::with_capacity_strategy(ITEM_COUNT, FindStrategy::PathSplitting)
+        );
+        test_with_different_types!(
+            path_compression,
+            path_compression_tester,
+            DisjointSets::with_capacity_strategy(ITEM_COUNT, FindStrategy::PathSplitting)
+        );
+        test_with_different_types!(
+            union_all,
+            union_all_tester,
+            DisjointSets::with_capacity_strategy(ITEM_COUNT, FindStrategy::PathSplitting)
+        );
+    }
 }
