@@ -2,6 +2,6 @@
 //! various traits and structs you will need.
 
 pub use crate::{
-    disjoint_sets::DisjointSets,
+    disjoint_sets::{DisjointSets, FindStrategy},
     traits::{Error, Result, UnionFind},
 };
